#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include <getopt.h>  /* getopt_long() */

#include <fcntl.h>  /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include <math.h>

#include <errno.h>

#include <limits.h>

#include <asm/types.h>  /* for videodev2.h */

#include <linux/videodev2.h>

#include <pthread.h>

#include "image_io.h"
#include "tools.h"
//#include "defs.h" /* Things needed in several modules */

#define CLEAR(x) memset(&(x), 0, sizeof(x))

#define NONE -1
#define IMG 0
#define MIN_HOLD 1
#define PEAK_HOLD 2
#define AVE 3
#define DIF 4
#define HIST 5
#define STRETCH 6
#define TRIG 7
#define TRIG_PEAK 8
#define REFERENCE 9
#define HISTOGRAM 10
#define DAY_DIR 11
#define MD_LOG_NAME 12
#define MD_LOG_DATA 13
#define STACK_LOG_NAME 14
#define STACK_LOG_DATA 15

/* Minimum delay between motion detection events in seconds */
#define DELAY_BETWEEN_MD 2

/* Struct definitions */
struct buffer {
  void *start;
  size_t length;
};

struct img_stack {
  unsigned long *Y;
  unsigned long *Cr; // half of size(Y)
  unsigned long *Cb; // half of size(Y)
  int width;
  int height;
  int channels;
  time_t timestamp;
  long usec;
  int num_imgs;
  unsigned long min_val;
  unsigned long max_val;
};


/* V4L2 format specifier */
struct v4l2_format fmt;

/* Image frames */

/* Current unpacked frame */
struct frame current_frame;

/* Ring memory for triggered capture */
struct frame *md_frames = NULL;
/* Peak hold stack of ring memory frames */
struct frame md_peak_hold;

/* Minimum stack */
struct frame *min_hold; /* Pointer to the stack in use */
struct frame *min_hold_save; /* Pointer to the stack being saved */
struct frame min_hold1; /* Data array 1 */
struct frame min_hold2; /* Data array 2 */
/* Maximum stack */
struct frame *peak_hold; /* Pointer to the stack in use */
struct frame *peak_hold_save; /* Pointer to the stack being saved */
struct frame peak_hold1; /* Data array 1 */
struct frame peak_hold2; /* Data array 2 */
/* Average stack */
/* This is the only stack that hold RGB values */
struct img_stack *ave_stack; /* Pointer to the stack in use */
struct img_stack *ave_stack_save; /* Pointer to the stack being saved */
struct img_stack ave_stack1; /* Data array 1 */
struct img_stack ave_stack2; /* Data array 2 */
/* Stacks that are used to save "latest.png" image, which is always of type PEAK_HOLD */
struct frame *latest_img; /* Pointer to the stack in use */
struct frame *latest_img_save; /* Pointer to the stack being saved */
struct frame latest_img1; /* Data array 1 */
struct frame latest_img2; /* Data array 2 */

/* Arrays for median values of the ring buffer pixels*/
unsigned char *md_median = NULL; /* Pointer to array in use */
unsigned char *md_median_calc = NULL; /* Pointer to array being calculated */
unsigned char *md_median1 = NULL; /* Data array 1 */
unsigned char *md_median2 = NULL; /* Data array 2 */

/* Arrays for detection size histograms */
int *md_ts_histogram = NULL; /* Array used in motion detection */
int *md_ts_histogram_calc = NULL; /* Array used in MD reference calculations */
int *md_ts_histogram_save = NULL;

/* Array for ts histogram of current frame */
int *frame_hist = NULL;
int *frame_hist_save = NULL;

/* Color conversion look-up tables from YUYV -> RGB */
unsigned char **yuyv2r = NULL; /* R channel */
unsigned char ***yuyv2g = NULL; /* G channel */
unsigned char **yuyv2b = NULL; /* B channel */

/* Luminances of each frame used in motion detection reference
   calculations */
unsigned char **md_luminance_calc = NULL;
/* Number of frames alread in md_luminance_calc array */
int frames_in_md_luminance_calc = 0;

static int img_in_mmap_buffer = 2;

/* Number of images used for each stack */
static int num_img_in_stack = 0;
/* Number of images in stacks being saved */
static int num_img_in_save_stack = 0;

/* Length of peak-hold/average/min stack in seconds */
static int stack_length = 60;

/* Length of "latest.png" stack in seconds */
static int latest_length = 60;

/* Number of image channels to be used (1 = B&W or 3 = RGB) */
static int number_of_channels = 1;

/* FPS requested from video device */
unsigned int video_input_fps = 25;

/* How many frames should be discarded before start of imaging */
unsigned int delay_start = 60;

static char *dev_name = NULL;
static int fd = -1;

/* MMAP buffers */
struct buffer *mmap_buffers = NULL;
static unsigned int n_mmap_buffers = 0;

/* Which type of images will be saved */
static int use_min = 0;
static int use_max = 0;
static int use_ave = 0;
static int use_dif = 0;
static int use_trig = 0;
static int use_latest = 0;

static int verbose = 0;

/* Thread for saving the ring memory */
pthread_t trig_save_thread;
/* Thread for saving stacks */
pthread_t stack_save_thread;
/* Thread for saving "latest.png" */
pthread_t latest_save_thread;
/* Thread for processing motion detection */
pthread_t md_trig_thread;
/* Thread for motion detection threshold calculations */
pthread_t md_references_thread;

/* Thread attributes for joinable threads */
pthread_attr_t thread_attr_join;
/* Thread attributes for detached threads */
pthread_attr_t thread_attr_detached;

/* Motion detection ring memory size */
static int img_in_md_buffer = 25;

pthread_t stack_thread, trig_thread;

/* Index of ring memory slot currently being used */
static int md_i = 0;

/* Amount of histogram adjustment in x and y directions */
static unsigned char md_histogram_adj = 10;

/* XXXXX Minimum change level allowed for detection */
unsigned char md_min_tl = 30;

/* How often the motion detection should be updated, in seconds */
static int md_threshold_update_interval = 60;

/* When was the md_background data updated */
static time_t md_background_update_time = 0;

/* If set, only print the motion detection result, do not save any images */
int md_test = 0;

/* Filename of a mask that can be used to exclude pixels from motion detection */
char *md_mask_fname = NULL;

/* Array used as user-selected mask of pixels to exclude from motion detection 
 */
unsigned char *md_mask = NULL;
int md_mask_width = -1;
int md_mask_height = -1;

/* Holds the index where detection happened. If no detection, value is -1. Value > -1 disables motion detection.
 */
int md_detection = -1;

/* Calculate how many images ago the triggering happened */
int md_images_since_detection = -1;

/* Latest triggered event */
time_t md_last_detection = 0;

/* Number of images to be saved before the event that triggered motion
   detection. Default [(img_in_md_buffer-1)/2] is set after buffer
   size is determined, unless user has set the value. */
int md_images_before = -1;

/* Date of the frame triggering the motion detection
*/
int md_trig_year = 0;
int md_trig_month = 0;
int md_trig_day = 0;

/* Number of motion detection events for the day
 */
int md_trig_number = 0;

/* Zone number that caused the triggering
*/
int md_triggered_zone = -1;

/* Time when the imaging started */
static time_t start_time;

/* Length of this imaging session */
static unsigned long capture_length = ULONG_MAX;

/* Starting time of the stacks */
static time_t stack_timestamp;
static long stack_usec;

/* Starting time of "latest.png" */
static time_t latest_img_timestamp;
static long latest_img_usec;

/* Ending time of the stacks */
static time_t stack_end_timestamp;
static long stack_end_usec;

/* Output directory path. '/' will be appended automatically */
static char *out_dir = NULL;

/* Prefix, which will be prepended to each filename. '_' will be added. */
static char *img_prefix = NULL;

/* Filename with full path for "latest.png" */
static char *latest_fname = NULL;
static char latest_fname_default[] = "latest.png";

/* Timestamp for forked process. Set to -1 if no child process exist */
time_t save_timestamp = -1;

/* Keep track if we are saving the motion detection ring memory */
int md_saving_images = 0;

/* Keep track if we are saving image stacks */
int saving_stacks = 0;

/* Keep track if we are saving "latest.png" */
int saving_latest = 0;

/* Keep track, if the md_process() thread is alive */
int md_process_on = 0;

/* Keep track if the ring memory is full, enable motion detection if
   it is. */
int md_ring_buffer_ready = 0;

/* Keep track if the calculation of motion detection references are
   being calculated */
int md_reference_calc_on = 0;

/* Motion detection references should be calculated */
int md_calc_references = 0;

/* Require a full buffer before calculating motion detection references */
int md_reference_buffer_ready = 0;

/* Set to 1, if motion detection reference buffers should be updated */
int md_update_reference_buffers = 0;

/* Require reference data before starting motion detection */
int md_references_ok = 0;

/* Log file to save daily data on triggered images */
char *md_log_fname = NULL;
FILE *md_log_fid = NULL;

/* Log file to save data on stacks */
char *stack_log_fname = NULL;
FILE *stack_log_fid = NULL;

/* Default to save both logs */
int save_log = 1;

/* Commandline options */
static const char short_options [] = "c:hn:s:mMadteT:H:U:k:GB:D:p:o:O:l:L:NC:f:V";

/* Long commandline options */
static const struct option long_options [] = {
  { "capture-device", required_argument, NULL, 'c' }, 
  { "buffer-size", required_argument, NULL, 'n' }, 
  { "stack-length", required_argument, NULL, 's' }, 
  { "min", no_argument, NULL, 'm' }, 
  { "max", no_argument, NULL, 'M' }, 
  { "ave", no_argument, NULL, 'a' }, 
  { "dif", no_argument, NULL, 'd' }, 
  { "trigger", no_argument, NULL, 't' },
  { "latest", no_argument, NULL, 'e' }, 
  { "minimum-trigger-level", required_argument, NULL, 'T' },
  { "histogram-adjust", required_argument, NULL, 'H' }, 
  { "update-interval", required_argument, NULL, 'U'}, 
  { "mask", required_argument, NULL, 'k' },
  { "trigger-test", no_argument, NULL, 'G' },
  { "images-before", required_argument, NULL, 'B' },
  { "delay-start", required_argument, NULL, 'D' }, 
  { "prefix", required_argument, NULL, 'p' }, 
  { "out-dir", required_argument, NULL, 'o' }, 
  { "latest-fname", required_argument, NULL, 'O' }, 
  { "capture-length", required_argument, NULL, 'l' }, 
  { "latest-length", required_argument, NULL, 'L' },
  { "no-log", no_argument, NULL, 'N' }, 
  { "number-of-channels", required_argument, NULL, 'C' }, 
  { "fps", required_argument, NULL, 'f' }, 
  { "verbose", no_argument, NULL, 'V' }, 
  { "help", no_argument, NULL, 'h' }, 
  { 0, 0, 0, 0 }
};



static void errno_exit(const char *s);

static int xioctl(int fd, int request, void *arg);

char *make_time_string(time_t timestamp, long usec, int type);

void make_fname(char *fname, time_t timestamp, long usec, int type);

void *md_save_images(void* ptr);

void *save_stacks(void *ptr);

void save_ave_stack(void);

void save_dif_stack(void);

void *md_process(void* ptr);

void *calc_md_references(void *ptr);

void *process_stack(void* ptr);

static int process_frame(const unsigned char *p, time_t timestamp, long usec);

static int read_frame(void);

static void mainloop(void);

static void stop_capturing(void);

static void start_capturing(void);

static void uninit_device(void);

static void init_mmap(int n);

static void init_device(int img_in_mmap_buffer);

static void close_device(void);

static void open_device(void);

static void usage (FILE *fp, int argc, char **argv);

void init_peak_hold(void);

void init_min_hold(void);

void init_latest_img(void);

void init_md_arrays(void);

void init_yuyv_lookups(void);

